"""Helper functions copied as-is from Werkzeug utils.

Ref: https://github.com/pallets/werkzeug/blob/master/src/werkzeug/utils.py

:copyright: 2007 Pallets
:license: BSD-3-Clause
"""

import pkgutil
from importlib import import_module


def find_modules(import_path, include_packages=False, recursive=False):
    """Finds all the modules below a package.

    This can be useful to
    automatically import all views / controllers so that their metaclasses /
    function decorators have a chance to register themselves on the
    application.

    Packages are not returned unless `include_packages` is `True`.  This can
    also recursively list modules but in that case it will import all the
    packages to get the correct load path of that module.

    :param import_path: the dotted name for the package to find child modules.
    :param include_packages: set to `True` if packages should be returned, too.
    :param recursive: set to `True` if recursion should happen.
    :return: generator
    """
    module = import_module(import_path)
    path = getattr(module, "__path__", None)
    if path is None:
        raise ValueError("%r is not a package" % import_path)
    basename = module.__name__ + "."
    for _importer, modname, ispkg in pkgutil.iter_modules(path):
        modname = basename + modname
        if ispkg:
            if include_packages:
                yield modname
            if recursive:
                for item in find_modules(modname, include_packages, True):
                    yield item
        else:
            yield modname



from flask import Flask

from deescovery import discover
from deescovery.flask import get_flask_rules


def test_discover_flask_should_load_blueprints(sample_project):
    flask_app = Flask("foo")
    rules = get_flask_rules("sample_project", flask_app)
    discover("sample_project", rules)
    assert list(flask_app.blueprints.keys()) == ["users"]


def test_discover_flask_should_load_cli(sample_project):
    flask_app = Flask("foo")
    rules = get_flask_rules("sample_project", flask_app)
    discover("sample_project", rules)
    assert list(flask_app.cli.commands.keys()) == ["users"]


def test_discover_flask_should_initialize_services(sample_project):
    flask_app = Flask("foo")
    rules = get_flask_rules("sample_project", flask_app)
    discover("sample_project", rules)
    from sample_project.services import bar, foo  # noqa

    assert foo.app == flask_app
    assert bar.app == flask_app



# Note: a number of type-ignore hints is due to an issue with mypy.
# Ref: https://github.com/python/mypy/issues/5485
import abc
import inspect
from dataclasses import dataclass
from importlib import import_module
from logging import getLogger
from typing import Any, Callable, List

from deescovery.contrib import find_modules

ModuleMatches = Callable[[str], bool]
ModuleAction = Callable[[str], Any]
ObjectMatches = Callable[[Any], bool]
ObjectAction = Callable[[Any], Any]


logger = getLogger(__name__)


class IRule(abc.ABC):
    """Generic type for a discovery rule.

    Normally, you use one of its subclasses:
    [ModuleRule][deescovery.discovery.ModuleRule] or
    [ObjectRule][deescovery.discovery.ObjectRule]
    """

    @abc.abstractmethod
    def discover(self, module_name: str) -> None:
        raise NotImplementedError()


@dataclass
class ModuleRule(IRule):
    """Do something with all modules matching the rule.

    For each module, found by [deescovery.discover][], the rule decides if it should
    be processed (calls `module_matcher`), and if the result is positive, calls
    `module_action`.

    **Example:**

    Import all `controllers.py` of the project.


    ```python
    from importlib import import_module
    from flask import Flask, Blueprint
    from deescovery import discover, ModuleRule
    from deescovery.matchers import MatchByPattern, MatchByType

    app = Flask(__name__)

    controller_loader = ModuleRule(
        name="Flask blueprints loader",
        module_matches=MatchByPattern(["*.controllers"]),
        module_action=import_module,
    )

    discover("myapp", [controller_loader])
    ```

    Attributes:
        name: the rule name. Used for logging purposes.
        module_matches: a callable (function) that takes the module name and returns
            True if the module should be processed. You can write your
            own ad-hoc function, or use a [deescovery.matchers.MatchByPattern][]
            class.

        module_action: a callable (function) that takes the module name. The action
            will only be executed if the module matches the pre-condition of
            `module_matches`. The most common action is importing the module to
            execute its content. For example, you can use it to register all API
            controllers that the module contains and defines with decorators.
    """

    name: str
    module_matches: ModuleMatches
    module_action: ModuleAction

    def discover(self, module_name: str) -> None:
        if self.module_matches(module_name):  # type: ignore
            logger.debug(f"{self.name} found module {module_name}")
            self.module_action(module_name)  # type: ignore


@dataclass
class ObjectRule(IRule):
    """Do something with all objects, matching the rules.

    For each module, found by [deescovery.discover][], the rule decides if the module
    should be processed (calls `module_matches`). If the result is positive, the module
    is imported, and all its members are inspected with `object_matches`. If
    the returned value of the object matcher is also True, calls `object_action`.

    **Example:**

    Find all [Flask blueprints](https://flask.palletsprojects.com/en/2.0.x/blueprints/)
    in the files `controllers.py` of the project and register them in the Flask
    application.

    ```python
    from flask import Flask, Blueprint
    from deescovery import discover, ObjectRule
    from deescovery.matchers import MatchByPattern, MatchByType

    app = Flask(__name__)

    blueprints_loader = ObjectRule(
        name="Flask blueprints loader",
        module_matches=MatchByPattern(["*.controllers"]),
        object_matches=MatchByType(Blueprint),
        object_action=app.register_blueprint,
    )

    discover("myapp", [blueprints_loader])
    ```

    Attributes:
        name: the rule name. Used for logging purposes.
        module_matches: a callable (function) that takes the module name and returns
            True if objects of the module should be inspected further with
            `object_matches`. You can write your own ad-hoc function or use a
            [deescovery.matchers.MatchByPattern][] class.
        object_matches: a callable (function) that takes the object and returns
            True if the object should be processed with `object_action`. You can write
            your own ad-hoc function or use one of the classes defined in the
            [deescovery.matchers][] module.
        object_action: a callable (function) that takes the found object.
            The action will only be executed when the object's module pre-condition
            passes the `module_matches` test, and object itself passes the
            pre-condition of `object_matches`.
    """

    name: str
    module_matches: ModuleMatches
    object_matches: ObjectMatches
    object_action: ObjectAction

    def discover(self, module_name: str) -> None:
        if not self.module_matches(module_name):  # type: ignore
            return
        module_obj = import_module(module_name)
        for object_name, obj in inspect.getmembers(
            module_obj, predicate=self.object_matches  # type: ignore
        ):
            logger.debug(f"{self.name} found {object_name} in {module_name}")
            self.object_action(obj)  # type: ignore


def discover(import_path: str, rules: List[IRule]):
    """Discover all objects.

    Scan the package, find all modules and objects, matching the given set of rules,
    and apply actions defined in them.

    Args:
        import_path: top-level module name to start scanning. Usually, it's a name of
            your application, e.g., "myapp". If your application doesn't have a single
            top-level module, you will probably call it for all top-level modules.
        rules: a list of module and objects rules. Each rule contains the
            match specification and the action, if the object matches.  Normally, it's
            a list IRule subclasses: [ModuleRule][deescovery.discovery.ModuleRule] or
            [ObjectRule][deescovery.discovery.ObjectRule].
    """
    for module_name in find_modules(import_path=import_path, recursive=True):
        for rule in rules:
            rule.discover(module_name)



import pathlib
from typing import List

from flask import Blueprint

from deescovery import ModuleRule, ObjectRule, discover
from deescovery.matchers import MatchByPattern, MatchByType


def test_module_rule_should_find_modules(collector: List, sample_project: pathlib.Path):
    rule = ModuleRule(
        name="Find controllers",
        module_matches=MatchByPattern(["*.controllers"]),
        module_action=collector.append,
    )
    discover("sample_project", rules=[rule])
    assert collector == ["sample_project.users.controllers"]


def test_object_module_rule_should_find_objects(
    collector: List, sample_project: pathlib.Path
):
    rule = ObjectRule(
        name="Find blueprints",
        module_matches=MatchByPattern(["*.controllers"]),
        object_matches=MatchByType(Blueprint),
        object_action=collector.append,
    )
    discover("sample_project", rules=[rule])
    assert len(collector) == 1
    assert collector[0].name == "users"



import pytest

from deescovery.matchers import MatchByMethod, MatchByPattern, MatchBySubclass


@pytest.mark.parametrize(
    "patterns, package, result",
    [
        (["*.models"], "domain_package.models", True),
        (["*.models_*"], "domain_package.models", False),
        (["*.models_*"], "domain_package.models_user", True),
        (["*.models.*"], "domain_package.models", False),
        (["*.models", "*.models.*"], "domain_package.models", True),
    ],
)
def test_match_by_pattern_should_return_relevant_packages(patterns, package, result):
    matcher = MatchByPattern(patterns)
    assert matcher(package) == result


@pytest.mark.parametrize(
    "attribute, obj, result",
    [
        ("strip", "", True),  # instances match
        ("strip", str, False),  # classes don't match
        ("foo", "", False),  # non-existent attribute
    ],
)
def test_match_by_method_should_return_relevant_objects(attribute, obj, result):
    matcher = MatchByMethod(attribute)
    assert matcher(obj) == result


class StringSubclass(str):
    pass


@pytest.mark.parametrize(
    "obj, result",
    [
        # instances don't match
        (1, False),
        ("foo", False),
        # the class itself doesn't match
        (str, False),
        # subclasses match
        (StringSubclass, True),
    ],
)
def test_match_by_subclass_should_return_relevant_objects(obj, result):
    matcher = MatchBySubclass(str)
    assert matcher(obj) == result



"""Module and object matchers."""
import fnmatch
import inspect
from dataclasses import dataclass
from typing import Any, List, Type


@dataclass
class MatchByPattern:
    """**Module matcher** that selects module names by patterns.

    **Example:**

    ```python
    matcher = MatchByPattern(["*.models", "*.models.*"])
    ```

    Attributes:
        patterns: the list of Unix shell-style wildcards for module names. E.g.
            the following instance will match all files `models.py` and
            `models/<something>.py` in a flat list of packages inside your application.
    """

    patterns: List[str]

    def __call__(self, value: str) -> bool:
        for pattern in self.patterns:
            if fnmatch.fnmatch(value, pattern):
                return True
        return False


@dataclass
class MatchByType:
    """**Object matcher** that selects instances by their type.

    Same as `lambda obj: isintance(obj, object_type)`.

    **Example:**

    Find all Flask blueprints in a module.

    ```python
    from flask import Blueprint
    matcher = MatchByType(Blueprint)
    ```

    Attributes:
        object_type: object type or a list of types.
    """

    object_type: Type

    def __call__(self, obj: Any):
        return isinstance(obj, self.object_type)


@dataclass
class MatchBySubclass:
    """**Object matcher** that select classes that are subclasses of a given type.

    Almost the same as `lambda obj: issubclass(obj, object_type)`.

    **Example:**

    Find all Django models.

    ```python
    from django.db import models
    matcher = MatchBySubclass(models.Model)
    ```

    Attributes:
        object_type: a type or a tuple of types.
    """

    object_type: Type

    def __call__(self, obj: Any):
        return (
            inspect.isclass(obj)
            and issubclass(obj, self.object_type)
            and obj != self.object_type
        )


@dataclass
class MatchByAttribute:
    """**Object matcher** that selects objects having an attribute with the given name.

    The same as `lambda obj: hasattr(obj, attribute_name)`

    **Example:**

    Find all objects that have an attribute `init_app`.

    ```python
    MatchByAttribute("init_app")
    ```

    Attributes:
        attribute_name: attribute name as a string.
    """

    attribute_name: str

    def __call__(self, obj: Any):
        return hasattr(obj, self.attribute_name)


@dataclass
class MatchByMethod:
    """**Object matcher** that selects objects having a method with a specific name.

    **Example:**

    Find all objects having a method `init_app()` (a common way for initializing Flask
    plugins.)

    ```python
    MatchByMethod("init_app")
    ```

    Attributes:
        method_name: method name as a string.
    """

    method_name: str

    def __call__(self, obj: Any):
        return (
            not inspect.isclass(obj)
            and hasattr(obj, self.method_name)
            and callable(getattr(obj, self.method_name))
        )
